_/*********************************************/
/*      LS & TS Device Driver Program       */
/*********************************************/

#define __KERNEL__
#define MODULE

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

#include "ls_ts.h"

static unsigned int busy = 0;



static int ls_ts_open(struct inode *inode, struct file *filp){
  if (busy) return - EBUSY;
  
  printk("ls_ts_open:\n");
  printk("   filp->f_version     : %lu \n",filp->f_version);
  
  linecount = LINECOUNT;
  printk("   LINECOUNT: %d \n",linecount);

  busy = 1;
  MOD_INC_USE_COUNT;

  ls_ts_detect(LS_TS_PORT_READ,LS_TS_PORT_MASK);

  return 0;
}


static ssize_t ls_ts_read(struct file *file, char *buf, size_t count, loff_t *offset){

  char *str;
  int len = LEN;
  str = it;


  if (linecount == 0){
     return 0;
  }else{
     linecount--;
  }

  *str = *ls_ts_module_read(3); 

  copy_to_user(buf,str,len);

  printk("ls_ts_read:\n");
  printk("   file->f_version     : %lu \n",file->f_version);
  printk("   requested: %d bytes   returned: %d bytes\n",count,len);
  printk("   message  : %s\n",str);
  return len;

}


static int ls_ts_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg){
  char *dat;
  int setcode,i;

  dat = it;

  /* printk("lcd_ioctl:\n");*/

  switch(cmd){
     #ifdef LS_TS_DEBUG
        case LS_TS_HARDRESET:
	   mod_use_count_ = 1;
     #endif

     case LS_READ:
       /*printk("   case LS_READ :\n");*/
	*dat = *ls_ts_module_read(0);
        setcode = 16;
        for(i=0;i<6;i++){
	   put_user(*dat, (char *)arg);
	   printk("   data %c \n",dat[i]); 
	   dat++;arg++;
        }
	break;

     case TS_READ:
       /*printk("   case TS_READ :\n");*/
	*dat = *ls_ts_module_read(1);
	dat = dat + 5;
        setcode=16;
        for(i=0;i<6;i++){
	   put_user(*dat, (char *)arg);
	   /* printk("   data %c \n",dat[i]);*/
	   dat++;arg++;
        }
	break;

     case PO_READ:
       /*printk("   case LS_TS_READ :\n");*/
	*dat = *ls_ts_module_read(2);
	dat = dat + 12;
	setcode=16;
	/*printk("      data = %c\n",dat[0]);*/
	put_user(*dat,(char *)arg);	
	break;

     default:
	return -EINVAL;
  }
  return 0;

}


static int ls_ts_release(struct inode *inode, struct file *filp){
  busy = 0;
  MOD_DEC_USE_COUNT;

  printk("ls_ts_close:\n");
  printk("   filp->f_version     : %lu \n",filp->f_version);

  ls_ts_module_release(LS_TS_PORT_READ,LS_TS_PORT_MASK);

  return 0;
}


static struct file_operations ls_ts_fops = {
   owner:              THIS_MODULE,
   read:               ls_ts_read,
   write:              NULL,
   ioctl:              ls_ts_ioctl,
   open:               ls_ts_open,
   release:            ls_ts_release,
 };


static int ls_ts_detect(unsigned int port, unsigned int range){
  int err;

  printk("ls_ts_detect:\n");

  if ((err=check_region(port,range)) < 0 ){
    printk("   This ports <%x - %x> are using now!\n",port,range);
    printk("   check_region(port,range) : %d\n",err);
    /* return err; */
  }     /* using port now */
  
  request_region(port,range,"ls_ts");     /* always success */

  printk("   I/Oport from: %x   to: %x \n",port,range);

  return 0;
}


static char *ls_ts_module_read(int n){
  int data,data1;
  char *dat;
  dat = it;

  /* printk("ls_ts_module_read:\n");*/

  data1 = inw(LS_TS_PORT_READ);
/*  data1 = 0x70f3; */
  data = data1;

  /* printk("inw(LS_TS_PORT_READ) -> : %x\n",data);*/

  if(n == 2 || n == 3){
     data = data & 0x1000;
     if(data >= 0x1000){
	dat[12] = '1';
     }else{
	dat[12] = '0';
     }
     if(n == 3){
  	data = data1;	
     }
  }

  if(n == 1 || n == 3){
     data = data & 0x0fc00;
     /*    printk("data = %x\n",data);*/
     if(data >= 0x0800){
        dat[11] = '1'; data = data - 0x0800;
     }else{
        dat[11] = '0';
     }
     if(data >= 0x0400){
        dat[10] = '1'; data = data - 0x0400;
     }else{
        dat[10] = '0';
     }
     if(data >= 0x0200){
        dat[9] = '1'; data = data - 0x0200;
     }else{
        dat[9] = '0';
     }
     if(data >= 0x0100){
        dat[8] = '1'; data = data - 0x0100;
     }else{
        dat[8] = '0';
     }
     if(data >= 0x0080){
        dat[7] = '1'; data = data - 0x0080;
     }else{
        dat[7] = '0';
     }
     if(data >= 0x0040){
        dat[6] = '1';data = data - 0x0040;
     }else{
        dat[6] = '0';
     }
     if(n == 3){
	data = data1;
     }
  }

  if(n == 0 || n == 3){
     data = data & 0x00f;
     /* printk("data = %x\n",data);*/
     if(data >= 0x0020){
        dat[5] = '1'; data = data - 0x0020;
	/*printk("data = %x\n",data);*/

     }else{
        dat[5] = '0';
     }
     if(data >= 0x0010){
        dat[4] = '1'; data = data - 0x0010;
     }else{
        dat[4] = '0';
	}
     if(data >= 0x0008){
        dat[3] = '1'; data = data - 0x0008;
     }else{
        dat[3] = '0';
     }
     if(data >= 0x0004){
        dat[2] = '1'; data = data - 0x0004;
     }else{
        dat[2] = '0';
     }
     if(data >= 0x0002){
        dat[1] = '1'; data = data - 0x0002;
     }else{
        dat[1] = '0';
     }
     if(data >= 0x0001){
        dat[0] = '1'; data = data - 0x0001;
     }else{
        dat[0] = '0';
     }
  }


/*  for(i=0;i<13;i++){
     printk("   data %d  : '%c'\n",i,dat[i]);
  }
*/
  return (dat);
}


static void ls_ts_module_release(unsigned int port,unsigned int range){
  release_region(port,range);
  printk("ls_ts_module_release:\n");
  printk("   I/Oport from: %x   to: %x \n",port,range);
}


int init_module (void){
  int i;

  printk("install 'ls_ts' into major %d\n",LS_TS_MAJOR);
  i = register_chrdev (LS_TS_MAJOR, "ls_ts", &ls_ts_fops);

  if (i){
     printk("device registration error\n");
     return -EBUSY;
  }

  return 0;
} 


void cleanup_module (void){
  int i;

  printk("remove 'ls_ts' from major %d\n",LS_TS_MAJOR);
  i = unregister_chrdev (LS_TS_MAJOR,"ls_ts");

  if (i){
     printk ("unregister_chrdev failed\n");
  }

}


