//#define MODULE
//#define __KERNEL__
//#define __RTL__
#include <linux/errno.h>
#include <asm/io.h>
#include <unistd.h>
#include <math.h>
#include <rtl.h>
#include <time.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>
#include "pwm_enc.h"

#define TASK_CONTROL_FIFO_OFFSET  4
#define COUNTER_RESET  1

#define Vmin     1
#define vshift   25
#define Kpl    0.1
#define Kpr    0.1
/*
#define Vmin     100
#define vshift   0
#define Kpl    0.01
#define Kpr    0.01
*/

static pthread_t tasks[1];
static int fflag = 0;

void *thread_code(void *t)
{
	//int fifo3 = (int)(t + 1);
	int fifo3 = 3;
	int fifo4 = fifo3 + 1;
	struct pwm_speed_data pwm;
	struct count_data count;

	int pwm_data_l1 = 0, pwm_data_r1 = 0;
	int pwm_data_l2 = 0, pwm_data_r2 = 0;
	int v = 0, w , x = 0, y ;  /* variable for the calculation */
	int abs1, abs2;
	int ret;
	int err;
	while(1){
		y = 0; w = 0;
		ret = pthread_wait_np();

		while((err = rtf_get(fifo4, &pwm, sizeof(pwm))) == sizeof(pwm)){
			pwm_data_l1 = -pwm.l_speed;
			pwm_data_r1 = pwm.r_speed;
  			pwm_data_l2 = 0;
			pwm_data_r2 = 0;
		}
		switch(pwm.mode){
			case DUTY_MODE:
				//rtl_printf("DUTY\n");
				//rtl_printf("pwm_data_r1 = %d pwm_data_l1 = %d\n",pwm_data_r1,pwm_data_l1);
					outw(pwm_data_l1, PWM_LEFT_WRITE);
					outw(pwm_data_r1, PWM_RIGHT_WRITE);
/*				if (pwm_data_l1 == KEEP_SPEED)
					outw(pwm_data_r1, PWM_RIGHT_WRITE);
				else if (pwm_data_r1 == KEEP_SPEED)
					outw(pwm_data_l1, PWM_LEFT_WRITE);
				else{
					outw(pwm_data_l1, PWM_LEFT_WRITE);
					outw(pwm_data_r1, PWM_RIGHT_WRITE);
				}
*/
				break;
			case STOP_MODE:
				rtl_printf("in enc_module6\n");
				v = pwm_data_l1 - pwm_data_l2;
				if(v > Vmin)
					w = Kpl * v + vshift;
				else if(v < -Vmin)
					w = Kpl * v - vshift;
				w = (w > 95) ? 95 : w;       
				w = (w < -95) ? -95 : w;     
				w = (w >= 0) ? (w + 128) : -w;
				x = pwm_data_r1 + pwm_data_r2;
				if(x > Vmin)
					y = Kpr * x + vshift;
				else if(x < -Vmin)
					y = Kpr * x - vshift;
				y = (y > 95) ? 95 : y;
				y = (y < -95) ? -95 : y;
				y = (y >= 0) ? (y + 128) : -y;
				
				rtl_printf("%7d %7d %7d %7d\n", pwm_data_l1, pwm_data_l2, pwm_data_r1, pwm_data_r2);
				
				abs1 = (v >= 0) ? v : -v;
				abs2 = (x >= 0) ? x : -x;

				if(abs1 <= Vmin && abs2 <= Vmin)
					count.br = 1;
				else
					count.br = 0;

				//printk("w = %d ; y = %d\n",w,y); 
				
				outw(w, PWM_LEFT_WRITE);
				outw(y, PWM_RIGHT_WRITE);
				break;
			default:
				break;
		}
		count.left_count = inw(LEFT_COUNTER);
		outw(COUNTER_RESET, LEFT_COUNTER);
		count.right_count = inw(RIGHT_COUNTER);
		outw(COUNTER_RESET, RIGHT_COUNTER);
		if(((count.left_count >> 15) & 1) == 1)
			count.left_count = -1 * (0x10000 - count.left_count);
 		pwm_data_l2 += count.left_count;  
		if(((count.right_count >> 15) & 1) == 1)
			count.right_count = -1 * (0x10000 - count.right_count);
		pwm_data_r2 += count.right_count;
		count.left_sum = pwm_data_l2;
		count.right_sum = pwm_data_r2;
		//rtl_printf("%7d %7d\n",count.left_count, count.right_count);
		rtf_put(fifo3, &count, sizeof(count));
	 }
	 return 0;
}

int my_handler(unsigned int fifo)
{
	struct my_msg_struct msg;
	int err;

	while((err = rtf_get(COMMAND_FIFO, &msg, sizeof(msg))) == sizeof(msg)){
		rtf_put(msg.task + TASK_CONTROL_FIFO_OFFSET, &msg, sizeof(msg));
		//rtl_printf("FIFO handler: sending the \"%d\" command to task %d; period %d\n", msg.command, msg.task, msg.period);
			switch(msg.command){
				case START_TASK:
					outw(COUNTER_RESET, LEFT_COUNTER);
					outw(COUNTER_RESET, RIGHT_COUNTER);
					printk("period=%d\n",msg.period);
					pthread_make_periodic_np(tasks[0], gethrtime(), (long long)msg.period * (long long)1000);
					pthread_wakeup_np(tasks[0]);
					break;
				case STOP_TASK:	
					//rtl_printf("Task %d: executing the \"%d\" command to task %d; period %d\n", msg.task, msg.command, msg.task, msg.period);
					pthread_suspend_np(tasks[0]);
					break;
				case RESTART_TASK:
					outw(COUNTER_RESET, LEFT_COUNTER);
					outw(COUNTER_RESET, RIGHT_COUNTER);
					pthread_wakeup_np(tasks[0]);
					break;
				default:
					rtl_printf("RTL task: bad command\n");
					return 0;
			}
	}
	if(err != 0){
		return -EINVAL;
	}
	return 0;
}

/* #define DEBUG */
int init_module(void)
{
	pthread_attr_t attr;
	struct sched_param sched_param;
	int c[4];
	int ret;

	rtf_destroy(3);
	rtf_destroy(4);
	rtf_destroy(5);
	rtf_destroy(6);
	c[0] = rtf_create(3, 4000);
	c[1] = rtf_create(4, 4000);
	c[2] = rtf_create(5, 200);      /* input control channel */
	c[3] = rtf_create(6, 100);      /* input control channel */
	rtl_printf("Fifo return 1=%d 2=%d 3=%d\n",c[0],c[1],c[2]);

	pthread_attr_init(&attr);
	sched_param.sched_priority = 5;
	pthread_attr_setschedparam(&attr, &sched_param);
	ret = pthread_create(&tasks[0], &attr, thread_code, (void *)2);
	pthread_setfp_np (tasks[0],1);
	fflag = 1;

	rtf_create_handler(5, &my_handler);
	return 0;
}

void cleanup_module(void)
{
	outw(0, PWM_LEFT_WRITE);
	outw(0, PWM_RIGHT_WRITE);
#ifdef DEBUG
	rtl_printf("%d\n", rtf_destroy(3));
	rtl_printf("%d\n", rtf_destroy(4));
	rtl_printf("%d\n", rtf_destroy(5));
	rtl_printf("%d\n", rtf_destroy(6));
#else
	rtf_destroy(3);
	rtf_destroy(4); 
	rtf_destroy(5);
	rtf_destroy(6);
#endif
	if(fflag == 1){
	pthread_delete_np (tasks[0]);
	}
}
