#include <linux/errno.h>
#include <rtl.h>
#include <time.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>
#include <asm/io.h>
#include <rtl_core.h>
#include <rtl_time.h>
#include <linux/ioport.h>

#define FIFO_1 1
#define FIFO_2 2
#define USS_IRQ 5
//pthread_t task1;

hrtime_t r_time;
hrtime_t s_time;
int i;
int data;

unsigned int transmit(int i)
{
	outb(1, 0x132);
	s_time = gethrtime();
	outb(i, 0x130);
	rtl_hard_enable_irq (USS_IRQ); 
        return 0;
}

int command_handler(unsigned int cmd)
{
	int err;
 	while ((err = rtf_get(FIFO_1, &cmd, sizeof(int))) == sizeof(int)) {
		switch (cmd) {
			case 1:
				transmit(0);
				break;
			case 2:
				transmit(1);
				break;
			case 3:
				transmit(2);
				break;
			default:
				return -EINVAL;
			}
	}
	return 0;
}

unsigned int intr_handler(unsigned int irq, struct pt_regs *regs)
{
	outb(2, 0x132);
	r_time = gethrtime();
        data = ((int)((float)r_time - (float)s_time)) * 343.7 / (1000000 * 2);
        rtf_put(FIFO_2, &data, sizeof(data));
	outb(1, 0x132);
	rtl_hard_disable_irq (USS_IRQ);
        return 0;
}

int init_module(void)
{	
 	rtf_create(1, 4000);
	rtf_create(2, 4000);
	if( check_region(0x130,4) != 0){
	printk("I/O error.\n");
 	return -ENODEV;
	}
  	request_region(0x130,4,"USS");
        rtf_create_handler(FIFO_1, &command_handler);
 	if(rtl_request_irq (USS_IRQ, intr_handler) != 0) {
	release_region(0x130,4);
	printk("IRQ error.\n");
	return -ENODEV;
	}
	return 0;
}

void cleanup_module(void)
{
	release_region(0x130,4);
	rtf_destroy(1);
	rtf_destroy(2);
	rtl_free_irq (USS_IRQ); 
}

